#define MOZ_SOURCE_STAMP 91faf7ec36cd18a8ebdc0e2edac966e5bbb15da2
#define MOZ_SOURCE_REPO https://hg.mozilla.org/releases/mozilla-beta
#define MOZ_SOURCE_URL https://hg.mozilla.org/releases/mozilla-beta/rev/91faf7ec36cd18a8ebdc0e2edac966e5bbb15da2
